
// Matrix product

#define MAX_TEMPORARY_SIZE 262144

// Pseudocode (in HPJava):
//
//   void matmul(S res [[,]], T src1 [[,]], U src2 [[,]]) {
//   
//     Group p = res.grp() ;
//     Range x = res.rng(0), y = res.rng(1) ;
//   
//     const int bandWidth = MAX_TEMPORARY_SIZE / 
//                           sizeof(T) * x.volume() + sizeof(U) * y.volume()) ;
//   
//     Index i = new Index(x), j = new Index(y) ;
//   
//     on(p)
//       over(i)
//         over(j)
//           res [i, j] = 0 ;
//   
//     int base = 0, residue = z1.size() ;
//     while(residue) {
//       const int b = bandWidth < residue ? bandWidth : residue ;
//   
//       // Create temporary arrays
//   
//       T tmp1 [[ ,*]] = new [[x, b]] on p ;
//       U tmp2 [[*, ]] = new [[b, y]] on p ;
//   
//       // Copy sections of sources to temporaries
//   
//       const int ceil = base + b - 1 ;
//       Adlib.remap(tmp1, src1 [[:, base : ceil]]) ;
//       Adlib.remap(tmp2, src2 [[base : ceil, :]]) ;
//   
//       // Local multiplication
//   
//       on(p)
//         over(i)
//           over(j)
//             for(int k = 0 ; k < b ; k++)
//               res [i, j] += tmp1 [i, k] * tmp2 [k, j] ;
//   
//       base    += b ;
//       residue -= b ;
//     }
//   }

template<class S, class T, class U>
inline void matmul(DAD* res,  S* resDat,
                   DAD* src1, T* src1Dat, DAD* src2, U* src2Dat) {
  Group p = res->grp() ;
  Range x = res->rng(0), y = res->rng(1) ;
  Map   s = res->map(0), t = res->map(1) ;

  Range x1 = src1->rng(0) ;
  Range y2 = src2->rng(1) ;
  Range z1 = src1->rng(1), z2 = src2->rng(0) ;

  // Require:
  //   x1.size() == x.size(), y2.size() == y.size(), z1.size() == z2.size()

  const int bandWidth =
      MAX_TEMPORARY_SIZE / (sizeof(T) * x.volume() + sizeof(U) * y.volume()) ;

  LocBlocksIndex i(x), j(y) ;

  if(p.member()) {
    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
      const int m = i.count ;
      const int resStep0  = s.step(i.sub_stp) ;

      const int resOffset0 = s.offset(i) ;

      for(j.beginLocBlk() ; j.test() ; j.nextLocBlk()) {
	const int n = j.count ;
	const int resStep1  = t.step(j.sub_stp) ;

	const int resOffset1 = resOffset0 + t.offset(j) ;

	S* resBlock = resDat  + resOffset1 ;

	for(int i = 0 ; i < m ; i++) {
	  S* resPtr = resBlock  ;
	  for(int j = 0 ; j < n ; j++) {
	    *resPtr = 0 ;
	    resPtr += resStep1  ;
	  }
	  resBlock += resStep0  ;
	}
      }
    }
  }

  int base = 0, residue = z1.size() ;
  while(residue) {
    const int b = bandWidth < residue ? bandWidth : residue ;
    CollapsedRange zb(b) ;

    Map tmp1_maps [2], tmp2_maps [2], sec1_maps [2], sec2_maps [2] ;
    Map s1, t2 ;
    int t1, s2 ;

    // Set up array descriptors for temporaries

    int size ;

    // Array2<T> tmp1(x, zb, p) ;

    size = 1 ;
    tmp1_maps [0] = Map(x, size) ;
    size *= x.volume() ;
    tmp1_maps [1] = Map(zb, size) ;
    size *= zb.volume() ;

    DAD tmp1(2, p, tmp1_maps) ;
    T* tmp1Dat = new T [size] ;

    s1 = tmp1.map(0) ;
    t1 = tmp1.str(1) ;

    // Array2<U> tmp2(zb, y, p) ;

    size = 1 ;
    tmp2_maps [0] = Map(zb, size) ;
    size *= zb.volume() ;
    tmp2_maps [1] = Map(y, size) ;
    size *= y.volume() ;

    DAD tmp2(2, p, tmp2_maps) ;
    U* tmp2Dat = new U [size] ;

    s2 = tmp2.str(0) ;
    t2 = tmp2.map(1) ;

    // Set up array descriptors for sections

    // Section2<T> sec1 = src1.sect(x1, Trip(base, ceil)) ;

    sec1_maps [0] = src1->map(0) ;
    sec1_maps [1] = Map(z1.subrng(b, base), src1->str(1)) ;
    DAD sec1(2, src1->group, sec1_maps) ;

    // Section2<U> sec2 = src2.sect(Trip(base, ceil), y2)) ;

    sec2_maps [0] = Map(z2.subrng(b, base), src2->str(0)) ;
    sec2_maps [1] = src2->map(1) ;
    DAD sec2(2, src2->group, sec2_maps) ;

    // Copy sections of source to temporaries

    Remap remap1(&tmp1, &sec1, sizeof(T)) ;
    remap1.execute(tmp1Dat, src1Dat) ;

    Remap remap2(&tmp2, &sec2, sizeof(U)) ;
    remap2.execute(tmp2Dat, src2Dat) ;

    // Local multiplication

    if(p.member()) {

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
	const int m = i.count ;
        const int resStep0  = s.step(i.sub_stp) ;
        const int tmp1Step0 = s1.step(i.sub_stp) ;

	const int resOffset0 = s.offset(i) ;
	const int tmp1Offset = s1.offset(i) ;

	for(j.beginLocBlk() ; j.test() ; j.nextLocBlk()) {
	  const int n = j.count ;
          const int resStep1  = t.step(j.sub_stp) ;
          const int tmp2Step1 = t2.step(j.sub_stp) ;

	  const int resOffset1 = resOffset0 + t.offset(j) ;
	  const int tmp2Offset = t2.offset(j) ;

	  S* resBlock  = resDat  + resOffset1 ;
	  T* tmp1Block = tmp1Dat + tmp1Offset ;
	  U* tmp2Block = tmp2Dat + tmp2Offset ;

	  // Should really use an optimized sequential matrix multiplier here.

	  for(int i = 0 ; i < m ; i++) {
	    S* resPtr   = resBlock  ;
	    U* tmp2Ptr1 = tmp2Block ;
	    for(int j = 0 ; j < n ; j++) {
	      T* tmp1Ptr  = tmp1Block ;
	      U* tmp2Ptr2 = tmp2Ptr1 ;
	      for(int k = 0 ; k < b ; k++) {
		*resPtr  += *tmp1Ptr * *tmp2Ptr2 ;
		tmp1Ptr  += t1 ;
		tmp2Ptr2 += s2 ;
	      }
	      resPtr   += resStep1  ;
	      tmp2Ptr1 += tmp2Step1 ;
	    }
	    resBlock  += resStep0  ;
	    tmp1Block += tmp1Step0 ;
	  }
	}
      }
    }

    delete [] tmp2Dat ;
    delete [] tmp1Dat ;

    base    += b ;
    residue -= b ;
  }
}

void boolMatmul(DAD* res,  int* resDat,
                DAD* src1, int* src1Dat, DAD* src2, int* src2Dat) ;

// -------------------------------------------------------------
// Things to do
//

